import json
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from pandas import DataFrame

from ..core.workspaces import resolve_workspace
from ..helpers.folders import (
    create_folders_from_path_string,
    resolve_folder_from_id_to_path,
)
from ..items.data_pipelines import (
    create_data_pipeline,
    get_data_pipeline,
    get_data_pipeline_definition,
    list_data_pipelines,
    resolve_data_pipeline,
    update_data_pipeline_definition,
)
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import (
    extract_display_name_from_platform,
    extract_middle_path,
    list_paths_of_type,
    pack_item_definition,
    unpack_item_definition,
)

logger = get_logger(__name__)


def get_data_pipeline_config(
    workspace: str, data_pipeline: str
) -> Union[Dict[str, Any], None]:
    """
    Get a specific data_pipeline config from a workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        data_pipeline (str): The name or ID of the data_pipeline.

    Returns:
        (Union[Dict[str, Any], None]): The dict config from the data_pipeline.
    """
    item = data_pipeline
    item_data = get_data_pipeline(workspace, item, df=False)

    if item_data is None:
        return None

    else:
        config = {}
        config = config[item_data.get('displayName')] = {}

        config = {
            'id': item_data['id'],
            'description': item_data.get('description', None),
            'folder_id': ''
            if item_data.get('folderId') is None
            or pd.isna(item_data.get('folderId'))
            else item_data['folderId'],
        }

        return config


def get_all_data_pipelines_config(
    workspace: str,
) -> Union[Dict[str, Any], None]:
    """
    Get data_pipelines config from a workspace.

    Args:
        workspace (str): The name or ID from the workspace.

    Returns:
        (Union[Dict[str, Any], None]): The dict config of all data_pipelines in the workspace
    """
    items = list_data_pipelines(workspace, df=False)

    if items is None:
        return None

    config = {}

    for item in items:

        item_data = get_data_pipeline(workspace, item['id'], df=False)

        config[item['displayName']] = {
            'id': item['id'],
            'description': item.get('description', None),
            'folder_id': ''
            if item.get('folderId') is None or pd.isna(item.get('folderId'))
            else item['folderId'],
        }

    return config


def export_data_pipeline(
    workspace: str,
    data_pipeline: str,
    path: Union[str, Path],
) -> None:
    """
    Export a data_pipeline to path.

    Args:
        workspace (str): The name or ID of the workspace.
        data_pipeline (str): The name or ID of the data_pipeline.
        path (Union[str, Path]): The path to export to.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    item = get_data_pipeline(workspace_id, data_pipeline, df=False)
    try:
        folder_path = resolve_folder_from_id_to_path(
            workspace_id, item['folderId']
        )
    except:
        logger.info(
            f'{item["displayName"]}.DataPipeline is not inside a folder.'
        )
        folder_path = None

    if folder_path is None:
        item_path = Path(path) / (item['displayName'] + '.DataPipeline')
    else:
        item_path = (
            Path(path) / folder_path / (item['displayName'] + '.DataPipeline')
        )
    os.makedirs(item_path, exist_ok=True)

    definition = get_data_pipeline_definition(workspace_id, item['id'])
    if definition is None:
        return None

    unpack_item_definition(definition, item_path)

    logger.success(
        f'`{item["displayName"]}.DataPipeline` was exported to {item_path} successfully.'
    )
    return None


def export_all_data_pipelines(
    workspace: str,
    path: Union[str, Path],
) -> None:
    """
    Export a data_pipeline to path.

    Args:
        workspace (str): The name or ID of the workspace.
        path (Union[str, Path]): The path to export to.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    items = list_data_pipelines(workspace_id, df=False)
    if items is None:
        return None

    for item in items:
        try:
            folder_path = resolve_folder_from_id_to_path(
                workspace_id, item['folderId']
            )
        except:
            logger.info(
                f'{item["displayName"]}.DataPipeline is not inside a folder.'
            )
            folder_path = None

        if folder_path is None:
            item_path = Path(path) / (item['displayName'] + '.DataPipeline')
        else:
            item_path = (
                Path(path)
                / folder_path
                / (item['displayName'] + '.DataPipeline')
            )
        os.makedirs(item_path, exist_ok=True)

        definition = get_data_pipeline_definition(workspace_id, item['id'])
        if definition is None:
            return None

        unpack_item_definition(definition, item_path)

    logger.success(f'All data_pipelines were exported to {path} successfully.')
    return None


@df
def deploy_data_pipeline(
    workspace: str,
    path: str,
    start_path: Optional[str] = None,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Deploy a data_pipeline to workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The path to the data_pipeline.
        start_path (Optional[str]): The starting path for folder creation.
        description (Optional[str]): Description for the data_pipeline.
        df (Optional[bool]): If True, returns a DataFrame, otherwise returns a dictionary.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The deployed data_pipeline or None if deployment fails.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    display_name = extract_display_name_from_platform(path)
    if display_name is None:
        return None

    item_id = resolve_data_pipeline(workspace_id, display_name)

    folder_path_string = extract_middle_path(path, start_path=start_path)
    folder_id = create_folders_from_path_string(
        workspace_id, folder_path_string
    )

    item_definition = pack_item_definition(path)

    if item_id is None:
        return create_data_pipeline(
            workspace_id,
            display_name=display_name,
            item_definition=item_definition,
            description=description,
            folder=folder_id,
            df=False,
        )

    else:
        return update_data_pipeline_definition(
            workspace_id,
            item_id,
            item_definition=item_definition,
            df=False,
        )


def deploy_all_data_pipelines(
    workspace: str,
    path: str,
    start_path: Optional[str] = None,
) -> None:
    """
    Deploy all data_pipelines to workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The path to the data_pipelines.
        start_path (Optional[str]): The starting path for folder creation.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    data_pipelines_paths = list_paths_of_type(path, 'DataPipeline')

    for path_ in data_pipelines_paths:

        display_name = extract_display_name_from_platform(path_)
        if display_name is None:
            return None

        item_id = resolve_data_pipeline(workspace_id, display_name)

        folder_path_string = extract_middle_path(path_, start_path=start_path)
        folder_id = create_folders_from_path_string(
            workspace_id, folder_path_string
        )

        item_definition = pack_item_definition(path_)

        if item_id is None:
            create_data_pipeline(
                workspace_id,
                display_name=display_name,
                item_definition=item_definition,
                folder=folder_id,
                df=False,
            )

        else:
            update_data_pipeline_definition(
                workspace_id,
                item_id,
                item_definition=item_definition,
                df=False,
            )

    logger.success(
        f'All data_pipelines were deployed to workspace "{workspace}" successfully.'
    )
    return None


def extract_data_pipeline_variables(path: str) -> List[Dict[str, str]]:
    """
    Extract data pipeline variables from the `pipeline-content.json` file.

    Args:
        path (str): The path to the data pipeline.

    Returns:
        List[Dict[str, str]]: A list of dictionaries containing the extracted variables.
    """

    path = Path(path) / 'pipeline-content.json'

    with open(path, 'r') as f:
        content = json.load(f)

    activities = content['properties']['activities']

    variables = []

    for activity_index, activity in enumerate(activities):
        activity_name = activity['name']

        subactivities = activity['typeProperties']['activities']
        for subactivity_index, subactivity in enumerate(subactivities):
            subactivity_name = subactivity['name']
            properties = subactivity['typeProperties']

            source = properties['source']['datasetSettings']
            source_database = source['typeProperties']['database']
            source_connection = source['externalReferences']['connection']

            sink = properties['sink']['datasetSettings']['linkedService']
            sink_name = sink['name']
            sink_properties = sink['properties']['typeProperties']
            sink_workspace_id = sink_properties['workspaceId']
            sink_artifact_id = sink_properties['artifactId']

            variables.append(
                {
                    'activity_index': activity_index,
                    'activity_name': activity_name,
                    'subactivity_index': subactivity_index,
                    'subactivity_name': subactivity_name,
                    'source_database': source_database,
                    'source_connection': source_connection,
                    'sink_name': sink_name,
                    'sink_workspace_id': sink_workspace_id,
                    'sink_artifact_id': sink_artifact_id,
                }
            )

    return variables


def replace_data_pipeline_variables_with_placeholders(
    path: str, variables: List[Dict[str, str]]
) -> None:
    """
    Replace data pipeline variables with placeholders in the pipeline content JSON.

    Args:
        path (str): The path to the data pipeline.
        variables (List[Dict[str, str]]): The list of variables to replace.
    """

    path = Path(path) / 'pipeline-content.json'

    with open(path, 'r') as f:
        content = json.load(f)

    for variable in variables:
        # Use indexes to find correct variable - Does not assume unique names
        # This allows multiple activities/subactivities with the same name
        activity_idx = variable['activity_index']
        subactivity_idx = variable['subactivity_index']

        # Substitute just the values that need to be replaced with placeholders
        # Database
        content['properties']['activities'][activity_idx]['typeProperties'][
            'activities'
        ][subactivity_idx]['typeProperties']['source']['datasetSettings'][
            'typeProperties'
        ][
            'database'
        ] = f"#{{{variable['activity_name']}_{variable['subactivity_name']}_source_database}}#"

        # Connection
        content['properties']['activities'][activity_idx]['typeProperties'][
            'activities'
        ][subactivity_idx]['typeProperties']['source']['datasetSettings'][
            'externalReferences'
        ][
            'connection'
        ] = f"#{{{variable['activity_name']}_{variable['subactivity_name']}_source_connection}}#"

        # Workspace ID
        content['properties']['activities'][activity_idx]['typeProperties'][
            'activities'
        ][subactivity_idx]['typeProperties']['sink']['datasetSettings'][
            'linkedService'
        ][
            'properties'
        ][
            'typeProperties'
        ][
            'workspaceId'
        ] = f"#{{{variable['activity_name']}_{variable['subactivity_name']}_sink_workspace_id}}#"

        # Artifact ID
        content['properties']['activities'][activity_idx]['typeProperties'][
            'activities'
        ][subactivity_idx]['typeProperties']['sink']['datasetSettings'][
            'linkedService'
        ][
            'properties'
        ][
            'typeProperties'
        ][
            'artifactId'
        ] = f"#{{{variable['activity_name']}_{variable['subactivity_name']}_sink_artifact_id}}#"

    modified_content = json.dumps(content, indent=2)

    # Save the modified content back to the file
    with open(path, 'w') as file:
        file.write(modified_content)


def _create_data_pipeline_placeholder_mapping(
    variables: List[Dict[str, str]]
) -> dict:
    """
    Creates a mapping of placeholders to their real values based on the extracted variables.
    """
    placeholder_mapping = {}

    for variable in variables:
        activity_name = variable['activity_name']
        subactivity_name = variable['subactivity_name']

        # Create a unique placeholder for each variable
        placeholder_mapping[
            f'{activity_name}_{subactivity_name}_source_database'
        ] = variable['source_database']
        placeholder_mapping[
            f'{activity_name}_{subactivity_name}_source_connection'
        ] = variable['source_connection']
        placeholder_mapping[
            f'{activity_name}_{subactivity_name}_sink_workspace_id'
        ] = variable['sink_workspace_id']
        placeholder_mapping[
            f'{activity_name}_{subactivity_name}_sink_artifact_id'
        ] = variable['sink_artifact_id']

    return placeholder_mapping


def replace_data_pipeline_placeholders_with_variables(
    path: str, variables: List
) -> None:
    """
    Replace data pipeline placeholders with their corresponding variable values.

    Args:
        path (str): The path to the data pipeline.
        variables (List[Dict[str, str]]): The list of variables to replace.
    """

    path = Path(path) / 'pipeline-content.json'

    with open(path, 'r') as f:
        content_str = f.read()

    mappings = _create_data_pipeline_placeholder_mapping(variables)

    # Substitute each placeholder with the corresponding value
    for placeholder, value in mappings.items():
        placeholder_pattern = f'#{{{placeholder}}}#'
        content_str = content_str.replace(placeholder_pattern, value)

    # Save the modified content back to the file
    with open(path, 'w') as file:
        file.write(content_str)



import json
import os
import uuid
from pathlib import Path
from typing import Any, Dict, Optional, Union

from ..api.api import _base_api
from ..core.workspaces import resolve_workspace
from ..items.dataflows_gen1 import (
    get_dataflow_gen1,
    get_dataflow_gen1_definition,
    list_dataflows_gen1,
)
from ..utils.logging import get_logger
from ..utils.utils import (
    list_paths_of_type,
    load_and_sanitize,
    write_single_line_json,
)

logger = get_logger(__name__)


def get_dataflow_gen1_config(
    workspace: str, dataflow_gen1: str
) -> Union[Dict[str, Any], None]:
    """
    Get a specific dataflow_gen1 config from a workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        dataflow_gen1 (str): The name or ID of the dataflow_gen1.

    Returns:
        (Union[Dict[str, Any], None]): The dict config from the dataflow_gen1.
    """
    item = dataflow_gen1
    item_data = get_dataflow_gen1(workspace, item, df=False)

    if item_data is None:
        return None

    else:
        config = {}
        config = config[item_data.get('name')] = {}

        config = {
            'id': item_data['objectId'],
            'description': item_data.get('description', None),
            'folder_id': '',
        }

        return config


def get_all_dataflows_gen1_config(
    workspace: str,
) -> Union[Dict[str, Any], None]:
    """
    Get dataflows_gen1 config from a workspace.

    Args:
        workspace (str): The name or ID from the workspace.

    Returns:
        (Union[Dict[str, Any], None]): The dict config of all dataflows_gen1 in the workspace
    """
    items = list_dataflows_gen1(workspace, df=False)

    if items is None:
        return None

    config = {}

    for item in items:
        config[item['name']] = {
            'id': item['objectId'],
            'description': item.get('description', None),
            'folder_id': '',
        }

    return config


def export_dataflow_gen1(
    workspace: str,
    dataflow: str,
    path: str,
) -> None:
    """
    Export a dataflow from a workspace to a file.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The dataflow name or ID.
        path (str, optional): The path to the project folder.

    Examples:
        ```python
        export_dataflow_gen1('MyProjectWorkspace', 'SalesDataflowGen1', path='path/to/project')
        export_dataflow_gen1('123e4567-e89b-12d3-a456-426614174000', 'SalesDataflowGen1', path='path/to/project')
        ```
    """
    workspace_id = resolve_workspace(workspace)
    if not workspace_id:
        return None

    # Get the dataflow details
    dataflow_ = get_dataflow_gen1(workspace_id, dataflow)
    if not dataflow_:
        return None

    dataflow_id = dataflow_['objectId']
    dataflow_name = dataflow_['name']

    definition_response = get_dataflow_gen1_definition(
        workspace=workspace_id,
        dataflow=dataflow_id,
    )

    if not definition_response:
        return None

    dataflow_name = dataflow_['name']
    dataflow_path = Path(path) / dataflow_name + '.Dataflow'
    os.makedirs(dataflow_path, exist_ok=True)

    # Save the model as model.json inside the item folder in single-line format (Power BI portal format)
    model_json_path = dataflow_path / 'model.json'
    write_single_line_json(definition_response, model_json_path)

    logger.success(f'Exported dataflow {dataflow_name} to {dataflow_path}.')
    return None


def export_all_dataflows_gen1(
    workspace: str,
    path: str,
) -> None:
    """
    Export all dataflows gen1 from a workspace to a file.

    Args:
        workspace (str): The workspace name or ID.
        path (str): The path to the project folder.

    Examples:
        ```python
        export_all_dataflows_gen1('MyProjectWorkspace', path='path/to/project')
        export_all_dataflows_gen1('123e4567-e89b-12d3-a456-426614174000', path='path/to/project')
        ```
    """
    workspace_id = resolve_workspace(workspace)
    if not workspace_id:
        return None

    dataflows = list_dataflows_gen1(workspace_id, df=False)

    if not dataflows:
        return None
    else:
        for dataflow in dataflows:
            export_dataflow_gen1(
                workspace=workspace, dataflow=dataflow['objectId'], path=path
            )
        return None


def _serialize_dataflow_gen1_model(path: str) -> tuple[bytes, str]:
    """
    Prepares the body for a dataflow deployment by reading and serializing the model.json file.

    Args:
        path (str): The path to the directory containing the model.json file.

    Returns:
        tuple[bytes, str]: The serialized multipart body and the boundary string.

    Raises:
        UnicodeEncodeError: If there is an encoding issue with the JSON content.

    Examples:
        ```python
        _serialize_dataflow_gen1_model('path/to/MyDataflowGen1.Dataflow')
        ```
    """
    # Read and clean JSON using load_and_sanitize function
    df_json = load_and_sanitize(Path(path) / 'model.json')

    json_str = json.dumps(df_json, ensure_ascii=False, separators=(',', ':'))

    # Boundary setup
    boundary = uuid.uuid4().hex
    LF = '\r\n'

    # Serialized Json Body
    body = (
        f'--{boundary}{LF}'
        f'Content-Disposition: form-data; name="model.json"; filename="model.json"{LF}'
        f'Content-Type: application/json{LF}{LF}'
        f'{json_str}{LF}'
        f'--{boundary}--{LF}'
    )

    try:
        body.encode('utf-8')
    except UnicodeEncodeError as e:
        logger.error(f'Encoding error: {e}')
        raise
    return body.encode('utf-8'), boundary


def deploy_dataflow_gen1(workspace: str, path: str) -> Union[bool, None]:
    """
    Deploy a dataflow in a workspace from a model.json file

    Args:
        workspace (str): The workspace name or ID.
        path (str): Path to the model.json file for the dataflow.

    Returns:
        None

    Raises:
        Exception: If the API request fails or returns an error.

    Examples:
        ```python
        deploy_dataflow_gen1('MyProjectWorkspace', 'path/to/MyDataflowGen1.Dataflow')
        deploy_dataflow_gen1('123e4567-e89b-12d3-a456-426614174000', 'path/to/MyDataflowGen1.Dataflow')
        ```
    """
    # Read and clean JSON
    body, boundary = _serialize_dataflow_gen1_model(path)

    content_type = f'multipart/form-data; boundary={boundary}'

    params = {
        'datasetDisplayName': 'model.json',
        'nameConflict': 'Abort',
    }

    workspace_id = resolve_workspace(workspace)
    if not workspace_id:
        return None

    response = _base_api(
        audience='powerbi',
        endpoint=f'/groups/{workspace_id}/imports',
        content_type=content_type,
        credential_type='user',
        method='post',
        data=body,
        params=params,
        return_raw=True,
    )
    # Handle response
    if not response.status_code in (200, 202):
        logger.error(
            f'Error deploying the dataflow: {response.status_code} - {response.json().get("error", {})}'
        )
        return None
    logger.success(f'Dataflow deployed successfully.')
    return True


def deploy_all_dataflows_gen1(
    workspace: str,
    path: str,
    start_path: Optional[str] = None,
) -> None:
    """
    Deploy all dataflows_gen1 to workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The path to the dataflows_gen2.
        start_path (Optional[str]): The starting path for folder creation.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    dataflows_gen2_paths = list_paths_of_type(path, 'Dataflow')

    for path_ in dataflows_gen2_paths:

        deploy_dataflow_gen1(workspace_id, path_)

    logger.success(
        f'All dataflows_gen1 were deployed to workspace "{workspace}" successfully.'
    )
    return None



from typing import Any, Dict, List, Literal, Optional, Union

from pandas import DataFrame

from ..api.api import api_request
from ..core.folders import resolve_folder
from ..core.workspaces import resolve_workspace
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import is_valid_uuid

logger = get_logger(__name__)


@df
def list_dataflows_gen1(
    workspace: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Returns a list of Gen1 dataflows in a specified workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, List[Dict[str, Any]], None]): A list of Gen1 dataflows or a DataFrame if df is True.
    """
    return api_request(
        endpoint='/groups/' + resolve_workspace(workspace) + '/dataflows',
        audience='powerbi',
        method='get',
        support_pagination=True,
    )


def get_dataflow_gen1_id(
    workspace: str, dataflow_name: str
) -> Union[str, None]:
    """
    Retrieves the ID of a Gen1 dataflow by its name.

    Args:
        dataflow_name (str): The name of the dataflow.

    Returns:
        str | None: The ID of the dataflow if found, otherwise None.
    """
    dataflows = list_dataflows_gen1(
        workspace=resolve_workspace(workspace),
        df=False,
    )
    for _dataflow in dataflows:
        if _dataflow['displayName'] == dataflow_name:
            return _dataflow['id']
    logger.warning(
        f"Dataflow '{dataflow_name}' not found in workspace '{workspace}'."
    )
    return None


def resolve_dataflow_gen1(workspace: str, dataflow: str) -> Union[str, None]:
    """
    Resolves a dataflow name to its ID.

    Args:
        workspace (str): The name or ID of the workspace.
        dataflow (str): The name or ID of the dataflow.

    Returns:
        str | None: The ID of the dataflow if found, otherwise None.
    """
    if is_valid_uuid(dataflow):
        return dataflow
    else:
        return get_dataflow_gen1_id(workspace, dataflow)


@df
def get_dataflow_gen1(
    workspace: str,
    dataflow: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Get a Power BI dataflow.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The name of the dataflow.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The dataflow.

    Examples:
        ```python
        get_dataflow_gen1('MyProjectWorkspace', 'SalesDataflowGen1')
        get_dataflow_gen1('123e4567-e89b-12d3-a456-426614174000', 'SalesDataflowGen1')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    return api_request(
        endpoint='/groups/'
        + workspace_id
        + '/dataflows/'
        + resolve_dataflow_gen1(workspace_id, dataflow),
        method='get',
        audience='powerbi',
    )


def get_dataflow_gen1_definition(workspace: str, dataflow: str) -> dict | None:
    """
    Get the definition of a Power BI dataflow Gen1.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The dataflow name or ID.

    Returns:
        dict: The dataflow definition.

    Examples:
        ```python
        get_dataflow_gen1_definition('MyProjectWorkspace', 'SalesDataflowGen1')
        get_dataflow_gen1_definition('123e4567-e89b-12d3-a456-426614174000', 'SalesDataflowGen1')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen1(workspace_id, dataflow)

    return api_request(
        endpoint='/groups/'
        + workspace_id
        + '/dataflows/'
        + dataflow_id
        + '/getDefinition',
        method='post',
        support_lro=True,
        audience='powerbi',
    )


@df
def update_dataflow_gen1_definition(
    workspace: str, dataflow: str, item_definition: Dict[str, Any]
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates the definition of an existing dataflow in the specified workspace.
    If the dataflow does not exist, it returns None.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The name or ID of the dataflow to update.
        item_definition (Dict[str, Any]): The item_definition of the dataflow.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated dataflow details if successful, otherwise None.

    Examples:
        ```python
        update_dataflow_gen1_definition(
            workspace='MyProjectWorkspace',
            dataflow='SalesDataflowGen1',
            item_definition={...} # The definition of the dataflow
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen1(workspace_id, dataflow)

    params = {'updateMetadata': True}
    payload = {'definition': item_definition}

    return api_request(
        endpoint='/groups/' + workspace_id + '/dataflows/' + dataflow_id,
        method='patch',
        payload=payload,
        params=params,
        audience='powerbi',
    )


@df
def create_dataflow_gen1(
    workspace: str,
    display_name: str,
    item_definition: Dict[str, Any],
    *,
    description: Optional[str] = None,
    folder: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Creates a new dataflow in the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        display_name (str): The display name of the dataflow.
        description (str, optional): A description for the dataflow.
        folder (str, optional): The folder to create the dataflow in.
        item_definition (Dict[str, Any]): The definition of the dataflow.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The created dataflow details.

    Examples:
        ```python
        create_dataflow_gen1(
            workspace='MyProjectWorkspace',
            display_name='SalesDataflowGen1',
            item_definition={...},
            description='This is a sales dataflow',
            folder='SalesDataflowsFolder'
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    payload = {
        'displayName': display_name,
        'definition': item_definition,
    }

    if folder:
        folder_id = resolve_folder(workspace_id, folder)
        if folder_id:
            payload['folderId'] = folder_id

    if description:
        payload['description'] = description

    return api_request(
        endpoint='/groups/' + workspace_id + '/dataflows',
        method='post',
        payload=payload,
        audience='powerbi',
    )


@df
def update_dataflow_gen1(
    workspace: str,
    dataflow: str,
    *,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates the properties of the specified dataflow.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The name or ID of the dataflow to update.
        display_name (str, optional): The new display name for the dataflow.
        description (str, optional): The new description for the dataflow.
        df (bool, optional): Keyword-only. If True, returns a DataFrame with flattened keys. Defaults to False.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated dataflow details if successful, otherwise None.

    Examples:
        ```python
        update_dataflow_gen1('MyProjectWorkspace', 'SalesDataflowGen1', display_name='UpdatedSalesDataflowGen1')
        update_dataflow_gen1('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000', description='Updated description')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen1(workspace_id, dataflow)

    payload = {}

    if display_name:
        payload['displayName'] = display_name

    if description:
        payload['description'] = description

    return api_request(
        endpoint='/groups/' + workspace_id + '/dataflows/' + dataflow_id,
        method='patch',
        payload=payload,
        audience='powerbi',
    )


def delete_dataflow_gen1(workspace: str, dataflow: str) -> None:
    """
    Deletes a dataflow from the specified workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        dataflow (str): The name or ID of the dataflow to delete.

    Returns:
        None: If the dataflow is successfully deleted.

    Examples:
        ```python
        delete_dataflow_gen1('MyProjectWorkspace', 'SalesDataflowGen1')
        delete_dataflow_gen1('123e4567-e89b-12d3-a456-426614174000', 'SalesDataflowGen1')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen1(workspace_id, dataflow)

    return api_request(
        endpoint='/groups/' + workspace_id + '/dataflows/' + dataflow_id,
        method='delete',
        audience='powerbi',
    )


def takeover_dataflow_gen1(workspace: str, dataflow: str) -> Union[bool, None]:
    """
    Take over a dataflow in a workspace

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The dataflow name or ID.

    Examples:
        ```python
        takeover_dataflow_gen1('MyProjectWorkspace', 'SalesDataflowGen1')
        takeover_dataflow_gen1('123e4567-e89b-12d3-a456-426614174000', 'SalesDataflowGen1')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen1(workspace_id, dataflow)

    return api_request(
        endpoint='/groups/'
        + workspace_id
        + '/dataflows/'
        + dataflow_id
        + '/Default.Takeover',
        method='post',
        support_lro=True,
    )


def refresh_dataflow_gen1(
    workspace: str,
    dataflow: str,
    *,
    process_type: Optional[str] = 'default',
    notify_option: Literal[
        'MailOnFailure', 'NoNotification'
    ] = 'NoNotification',
) -> None:
    """
    Refresh a dataflow in the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The name or ID of the dataflow to refresh.
        process_type (str, optional): The process type to use for the refresh. Defaults to 'default'.
        notify_option (Literal['MailOnFailure', 'NoNotification'], optional): The notification option to use for the refresh. Defaults to 'NoNotification'.

    Returns:
        None: If the refresh was successful.

    Examples:
        ```python
        refresh_dataflow_gen1('MyProjectWorkspace', 'SalesDataflow')
        refresh_dataflow_gen1('123e4567-e89b-12d3-a456-426614174000', 'SalesDataflow')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen1(workspace_id, dataflow)

    payload = {'notifyOption': notify_option}

    params = {'processType': process_type}

    return api_request(
        endpoint='/groups/'
        + workspace_id
        + '/dataflows/'
        + dataflow_id
        + '/refreshes',
        method='post',
        payload=payload,
        params=params,
        audience='powerbi',
    )


@df
def get_dataflow_gen1_transactions(
    workspace: str,
    dataflow: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Get transactions for a dataflow in a workspace.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The dataflow name or ID.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        Union[DataFrame, List[Dict[str, Any]], None]: The dataflow transactions or None if not found.

    Examples:
        ```python
        get_dataflow_gen1_transactions('MyProjectWorkspace', 'SalesDataflowGen1')
        get_dataflow_gen1_transactions('123e4567-e89b-12d3-a456-426614174000', 'SalesDataflowGen1')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen1(workspace_id, dataflow)

    return api_request(
        endpoint='/groups/'
        + workspace_id
        + '/dataflows/'
        + dataflow_id
        + '/transactions',
        method='post',
        audience='powerbi',
    )


@df
def get_dataflows_gen1_datasources(
    workspace: str,
    dataflow: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Get the data sources for a dataflow in a workspace.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The dataflow name or ID.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        Union[DataFrame, List[Dict[str, Any]], None]: The dataflow datasources or None if not found.

    Examples:
        ```python
        get_dataflows_gen1_datasources('MyProjectWorkspace', 'SalesDataflowGen1')
        get_dataflows_gen1_datasources('123e4567-e89b-12d3-a456-426614174000', 'SalesDataflowGen1')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen1(workspace_id, dataflow)

    return api_request(
        endpoint='/groups/'
        + workspace_id
        + '/dataflows/'
        + dataflow_id
        + '/datasources',
        method='post',
        audience='powerbi',
    )



from typing import Any, Dict, List, Optional, Union

from pandas import DataFrame

from ..api.api import api_request
from ..core.folders import resolve_folder
from ..core.workspaces import resolve_workspace
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import is_valid_uuid

logger = get_logger(__name__)


@df
def list_dataflows_gen2(
    workspace: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Lists all dataflows in a workspace.

    Args:
        workspace (str): The workspace name or ID.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        list | pandas.DataFrame | None: A list of dataflows if successful, otherwise None.
    """
    workspace_id = resolve_workspace(workspace)

    return api_request(
        endpoint='/workspaces/' + workspace_id + '/dataflows',
        support_pagination=True,
    )


def get_dataflow_gen2_id(
    workspace: str, dataflow_gen2_name: str
) -> Union[str, None]:
    """
    Retrieves the ID of a dataflow by its name.

    Args:
        dataflow_gen2_name (str): The name of the dataflow.

    Returns:
        (Union[str, None]): The ID of the dataflow if found, otherwise None.
    """
    dataflows = list_dataflows_gen2(
        workspace=resolve_workspace(workspace),
        df=False,
    )

    for _dataflow in dataflows:
        if _dataflow['displayName'] == dataflow_gen2_name:
            return _dataflow['id']
    logger.warning(
        f"Dataflow '{dataflow_gen2_name}' not found in workspace '{workspace}'."
    )
    return None


def resolve_dataflow_gen2(
    workspace: str,
    dataflow: str,
) -> Union[str, None]:
    """
    Resolves a dataflow name to its ID.

    Args:
        workspace (str): The ID of the workspace.
        dataflow (str): The name of the dataflow.

    Returns:
        (Union[str, None]): The ID of the dataflow, or None if not found.

    Examples:
        ```python
        resolve_dataflow('MyProjectWorkspace', 'SalesDataflow')
        resolve_dataflow('123e4567-e89b-12d3-a456-426614174000', 'SalesDataflow')
        ```
    """
    if is_valid_uuid(dataflow):
        return dataflow
    else:
        return get_dataflow_gen2_id(resolve_workspace(workspace), dataflow)


@df
def get_dataflow_gen2(
    workspace: str,
    dataflow: str,
    *,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Gets a dataflow by its name or ID.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The name or ID of the dataflow.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The dataflow details if found, otherwise None.

    Examples:
        ```python
        get_dataflow('MyProjectWorkspace', 'SalesDataflow')
        get_dataflow('123e4567-e89b-12d3-a456-426614174000', 'SalesDataflow')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen2(workspace_id, dataflow)

    return api_request(
        endpoint='/workspaces/' + workspace_id + '/dataflows/' + dataflow_id,
    )


@df
def update_dataflow_gen2(
    workspace: str,
    dataflow: str,
    *,
    display_name: str = None,
    description: str = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates the properties of the specified dataflow.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The name or ID of the dataflow to update.
        display_name (str, optional): The new display name for the dataflow.
        description (str, optional): The new description for the dataflow.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated dataflow details if successful, otherwise None.

    Examples:
        ```python
        update_dataflow('MyProjectWorkspace', 'SalesDataModel', display_name='UpdatedSalesDataModel')
        update_dataflow('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000', description='Updated description')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen2(workspace_id, dataflow)

    payload = {}

    if display_name:
        payload['displayName'] = display_name

    if description:
        payload['description'] = description

    return api_request(
        endpoint='/workspaces/' + workspace_id + '/dataflows/' + dataflow_id,
        method='patch',
        payload=payload,
    )


def delete_dataflow_gen2(workspace: str, dataflow: str) -> None:
    """
    Delete a dataflow from the specified workspace.

    Args:
        workspace (str): The name or ID of the workspace to delete.
        dataflow (str): The name or ID of the dataflow to delete.

    Returns:
        None: If the dataflow is successfully deleted.

    Raises:
        ResourceNotFoundError: If the specified workspace is not found.

    Examples:
        ```python
        delete_dataflow('MyProjectWorkspace', 'SalesDataflow')
        delete_dataflow('123e4567-e89b-12d3-a456-426614174000', 'SalesDataflow')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen2(workspace_id, dataflow)

    return api_request(
        endpoint='/workspaces/' + workspace_id + '/dataflows/' + dataflow_id,
        method='delete',
    )


def get_dataflow_gen2_definition(
    workspace: str, dataflow: str
) -> Union[Dict[str, Any], None]:
    """
    Retrieves the definition of a dataflow by its name or ID from the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The name or ID of the dataflow.

    Returns:
        (Union[Dict[str, Any], None]): The dataflow definition if found, otherwise None.

    Examples:
        ```python
        get_dataflow_definition('MyProjectWorkspace', 'Salesdataflow')
        get_dataflow_definition('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    # Resolving IDs
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen2(workspace_id, dataflow)

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/dataflows/'
        + dataflow_id
        + '/getDefinition',
        method='post',
        support_lro=True,
    )


@df
def update_dataflow_gen2_definition(
    workspace: str,
    dataflow: str,
    item_definition: Dict[str, Any],
    df: Optional[bool] = True,
) -> Union[Dict[str, Any], None]:
    """
    Updates the definition of an existing dataflow in the specified workspace.
    If the dataflow does not exist, it returns None.

    Args:
        workspace (str): The workspace name or ID.
        dataflow (str): The name or ID of the dataflow to update.
        item_definition (Dict[str, Any]): The updated item definition.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[Dict[str, Any], None]): The updated dataflow details if successful, otherwise None.

    Examples:
        ```python
        update_dataflow('MyProjectWorkspace', 'SalesDataModel', display_name='UpdatedSalesDataModel')
        update_dataflow('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000', description='Updated description')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    dataflow_id = resolve_dataflow_gen2(workspace_id, dataflow)

    payload = {'definition': item_definition}

    params = {'updateMetadata': True}

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/dataflows/'
        + dataflow_id
        + '/updateDefinition',
        method='patch',
        payload=payload,
        params=params,
        support_lro=True,
    )


@df
def create_dataflow_gen2(
    workspace: str,
    display_name: str,
    item_definition: Dict[str, Any],
    *,
    description: Optional[str] = None,
    folder: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[dict, None]:
    """
    Creates a new dataflow in the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        display_name (str): The display name of the dataflow.
        description (str, optional): A description for the dataflow.
        folder (str, optional): The folder to create the dataflow in.
        item_definition (Dict[str, Any]): The definition of the dataflow.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (dict): The created dataflow details.

    Examples:
        ```python
        create_dataflow('MyProjectWorkspace', 'SalesDataModel', 'path/to/definition.json')
        create_dataflow('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000', 'path/to/definition.json', description='Sales data model')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    payload = {'displayName': display_name, 'definition': item_definition}

    if description:
        payload['description'] = description

    if folder:
        folder_id = resolve_folder(workspace_id, folder)
        if folder_id:
            payload['folderId'] = folder_id

    return api_request(
        endpoint='/workspaces/' + workspace_id + '/dataflows',
        method='post',
        payload=payload,
        support_lro=True,
    )



import json
import os
from pathlib import Path
from typing import Any, Dict, Optional, Union

from pandas import DataFrame

from ..core.workspaces import resolve_workspace
from ..helpers.folders import (
    create_folders_from_path_string,
    resolve_folder_from_id_to_path,
)
from ..items.items import (
    create_item,
    get_item,
    get_item_definition,
    list_items,
    resolve_item,
    update_item_definition,
)
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import (
    extract_display_name_from_platform,
    extract_middle_path,
    list_paths_of_type,
    pack_item_definition,
    unpack_item_definition,
)

logger = get_logger(__name__)


def export_item(
    workspace: str,
    item: str,
    path: str,
):
    """
    Exports a item definition to a specified folder structure.

    Args:
        workspace (str): The workspace name or ID.
        item (str): The name of the item to export.
        path (str): The root path of the project.

    Examples:
        ```python
        export_item('MyProjectWorkspace', 'SalesDataModel', '/path/to/project')
        export_item('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000', '/path/to/project')
        ```
    """
    workspace_id = resolve_workspace(workspace)
    if not workspace_id:
        return None

    item_ = get_item(workspace_id, item, df=False)
    if not item_:
        return None

    item_id = item_['id']
    item_type = item_['type']
    folder_id = None
    if 'folderId' in item_:
        folder_id = item_['folderId']

    definition = get_item_definition(workspace_id, item_id)
    if not definition:
        return None

    try:
        folder_path = resolve_folder_from_id_to_path(workspace_id, folder_id)
    except:
        logger.info(
            f'{item["displayName"]}.{item_type} is not inside a folder.'
        )
        folder_path = None

    if folder_path is None:
        item_path = Path(path) / (item['displayName'] + f'.{item_type}')
    else:
        item_path = (
            Path(path) / folder_path / (item['displayName'] + f'.{item_type}')
        )
    os.makedirs(item_path, exist_ok=True)

    unpack_item_definition(definition, item_path)

    logger.success(
        f'`{item["displayName"]}.{item_type}` was exported to {item_path} successfully.'
    )
    return None


def export_all_items(
    workspace: str,
    path: str,
) -> None:
    """
    Exports all items to the specified folder structure.

    Args:
        workspace (str): The workspace name or ID.
        path (str): The root path of the project.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    items = list_items(workspace_id, df=False)
    if items is None:
        return None

    for item in items:
        item_ = get_item(workspace_id, item, df=False)
        if not item_:
            return None

        item_id = item_['id']
        item_type = item_['type']
        folder_id = None
        if 'folderId' in item_:
            folder_id = item_['folderId']

        definition = get_item_definition(workspace_id, item_id)
        if not definition:
            return None

        try:
            folder_path = resolve_folder_from_id_to_path(
                workspace_id, folder_id
            )
        except:
            logger.info(
                f'{item["displayName"]}.{item_type} is not inside a folder.'
            )
            folder_path = None

        if folder_path is None:
            item_path = Path(path) / (item['displayName'] + f'.{item_type}')
        else:
            item_path = (
                Path(path)
                / folder_path
                / (item['displayName'] + f'.{item_type}')
            )
        os.makedirs(item_path, exist_ok=True)

        unpack_item_definition(definition, item_path)

        logger.success(
            f'`{item["displayName"]}.{item_type}` was exported to {item_path} successfully.'
        )
    return None


@df
def deploy_item(
    workspace: str,
    path: str,
    start_path: Optional[str] = None,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Creates or updates a item in Fabric based on local folder structure.
    Automatically detects the folder_id based on where the item is located locally.

    Args:
        workspace (str): The workspace name or ID.
        path (str): The root path of the project.
        start_path (str, optional): The starting path for the item.
        description (str, optional): A description for the item.
        df (bool, optional): Whether to return a DataFrame. Defaults to True.
    """
    workspace_id = resolve_workspace(workspace)
    if not workspace_id:
        return None

    display_name = extract_display_name_from_platform(path)
    if display_name is None:
        return None

    item_id = resolve_item(workspace_id, display_name)

    folder_path_string = extract_middle_path(path, start_path=start_path)
    folder_id = create_folders_from_path_string(
        workspace_id, folder_path_string
    )

    item_definition = pack_item_definition(path)

    if item_id is None:
        return create_item(
            workspace_id,
            display_name=display_name,
            item_definition=item_definition,
            description=description,
            folder=folder_id,
            df=False,
        )

    else:
        return update_item_definition(
            workspace_id,
            item_id,
            item_definition=item_definition,
            df=False,
        )


def deploy_all_items(
    workspace: str,
    path: str,
    start_path: Optional[str] = None,
) -> None:
    """
    Deploy all items to workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The path to the notebooks.
        start_path (Optional[str]): The starting path for folder creation.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    types = ['Notebook', 'DataPipeline', 'Dataflow', 'SemanticModel', 'Report']
    for type in types:
        item_paths = list_paths_of_type(path, type)

        for path_ in item_paths:

            display_name = extract_display_name_from_platform(path_)
            if display_name is None:
                return None

            item_id = resolve_item(workspace_id, display_name)

            folder_path_string = extract_middle_path(
                path_, start_path=start_path
            )
            folder_id = create_folders_from_path_string(
                workspace_id, folder_path_string
            )

            item_definition = pack_item_definition(path_)

            if item_id is None:
                create_item(
                    workspace_id,
                    display_name=display_name,
                    item_definition=item_definition,
                    folder=folder_id,
                    df=False,
                )

            else:
                update_item_definition(
                    workspace_id,
                    item_id,
                    item_definition=item_definition,
                    df=False,
                )

    logger.success(
        f'All items were deployed to workspace "{workspace}" successfully.'
    )
    return None



import os
import re
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from pandas import DataFrame

from ..core.workspaces import resolve_workspace
from ..helpers.folders import (
    create_folders_from_path_string,
    resolve_folder_from_id_to_path,
)
from ..items.dataflows_gen2 import (
    create_dataflow_gen2,
    get_dataflow_gen2,
    get_dataflow_gen2_definition,
    list_dataflows_gen2,
    resolve_dataflow_gen2,
    update_dataflow_gen2_definition,
)
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import (
    extract_display_name_from_platform,
    extract_middle_path,
    list_paths_of_type,
    pack_item_definition,
    unpack_item_definition,
)

logger = get_logger(__name__)


def get_dataflow_gen2_config(
    workspace: str, dataflow_gen2: str
) -> Union[Dict[str, Any], None]:
    """
    Get a specific dataflow_gen2 config from a workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        dataflow_gen2 (str): The name or ID of the dataflow_gen2.

    Returns:
        (Union[Dict[str, Any], None]): The dict config from the dataflow_gen2.
    """
    item = dataflow_gen2
    item_data = get_dataflow_gen2(workspace, item, df=False)

    if item_data is None:
        return None

    else:
        config = {}
        config = config[item_data.get('displayName')] = {}

        config = {
            'id': item_data['id'],
            'description': item_data.get('description', None),
            'folder_id': ''
            if item_data.get('folderId') is None
            or pd.isna(item_data.get('folderId'))
            else item_data['folderId'],
        }

        return config


def get_all_dataflows_gen2_config(
    workspace: str,
) -> Union[Dict[str, Any], None]:
    """
    Get dataflows_gen2 config from a workspace.

    Args:
        workspace (str): The name or ID from the workspace.

    Returns:
        (Union[Dict[str, Any], None]): The dict config of all dataflows_gen2 in the workspace
    """
    items = list_dataflows_gen2(workspace, df=False)

    if items is None:
        return None

    config = {}

    for item in items:

        item_data = get_dataflow_gen2(workspace, item['id'], df=False)

        config[item['displayName']] = {
            'id': item['id'],
            'description': item.get('description', None),
            'folder_id': ''
            if item.get('folderId') is None or pd.isna(item.get('folderId'))
            else item['folderId'],
        }

    return config


def export_dataflow_gen2(
    workspace: str,
    dataflow_gen2: str,
    path: Union[str, Path],
) -> None:
    """
    Export a dataflow_gen2 to path.

    Args:
        workspace (str): The name or ID of the workspace.
        dataflow_gen2 (str): The name or ID of the dataflow_gen2.
        path (Union[str, Path]): The path to export to.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    item = get_dataflow_gen2(workspace_id, dataflow_gen2, df=False)
    try:
        folder_path = resolve_folder_from_id_to_path(
            workspace_id, item['folderId']
        )
    except:
        logger.info(f'{item["displayName"]}.Dataflow is not inside a folder.')
        folder_path = None

    if folder_path is None:
        item_path = Path(path) / (item['displayName'] + '.Dataflow')
    else:
        item_path = (
            Path(path) / folder_path / (item['displayName'] + '.Dataflow')
        )
    os.makedirs(item_path, exist_ok=True)

    definition = get_dataflow_gen2_definition(workspace_id, item['id'])
    if definition is None:
        return None

    unpack_item_definition(definition, item_path)

    logger.success(
        f'`{item["displayName"]}.Dataflow` was exported to {item_path} successfully.'
    )
    return None


def export_all_dataflows_gen2(
    workspace: str,
    path: Union[str, Path],
) -> None:
    """
    Export a dataflow_gen2 to path.

    Args:
        workspace (str): The name or ID of the workspace.
        path (Union[str, Path]): The path to export to.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    items = list_dataflows_gen2(workspace_id, df=False)
    if items is None:
        return None

    for item in items:
        try:
            folder_path = resolve_folder_from_id_to_path(
                workspace_id, item['folderId']
            )
        except:
            logger.info(
                f'{item["displayName"]}.Dataflow is not inside a folder.'
            )
            folder_path = None

        if folder_path is None:
            item_path = Path(path) / (item['displayName'] + '.Dataflow')
        else:
            item_path = (
                Path(path) / folder_path / (item['displayName'] + '.Dataflow')
            )
        os.makedirs(item_path, exist_ok=True)

        definition = get_dataflow_gen2_definition(workspace_id, item['id'])
        if definition is None:
            return None

        unpack_item_definition(definition, item_path)

    logger.success(f'All dataflows_gen2 were exported to {path} successfully.')
    return None


@df
def deploy_dataflow_gen2(
    workspace: str,
    path: str,
    start_path: Optional[str] = None,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Deploy a dataflow_gen2 to workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The path to the dataflow_gen2.
        start_path (Optional[str]): The starting path for folder creation.
        description (Optional[str]): Description for the dataflow_gen2.
        df (Optional[bool]): If True, returns a DataFrame, otherwise returns a dictionary.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The deployed dataflow_gen2 or None if deployment fails.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    display_name = extract_display_name_from_platform(path)
    if display_name is None:
        return None

    item_id = resolve_dataflow_gen2(workspace_id, display_name)

    folder_path_string = extract_middle_path(path, start_path=start_path)
    folder_id = create_folders_from_path_string(
        workspace_id, folder_path_string
    )

    item_definition = pack_item_definition(path)

    if item_id is None:
        return create_dataflow_gen2(
            workspace_id,
            display_name=display_name,
            item_definition=item_definition,
            description=description,
            folder=folder_id,
            df=False,
        )

    else:
        return update_dataflow_gen2_definition(
            workspace_id,
            item_id,
            item_definition=item_definition,
            df=False,
        )


def deploy_all_dataflows_gen2(
    workspace: str,
    path: str,
    start_path: Optional[str] = None,
) -> None:
    """
    Deploy all dataflows_gen2 to workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        path (str): The path to the dataflows_gen2.
        start_path (Optional[str]): The starting path for folder creation.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    dataflows_gen2_paths = list_paths_of_type(path, 'Dataflow')

    for path_ in dataflows_gen2_paths:

        display_name = extract_display_name_from_platform(path_)
        if display_name is None:
            return None

        item_id = resolve_dataflow_gen2(workspace_id, display_name)

        folder_path_string = extract_middle_path(path_, start_path=start_path)
        folder_id = create_folders_from_path_string(
            workspace_id, folder_path_string
        )

        item_definition = pack_item_definition(path_)

        if item_id is None:
            create_dataflow_gen2(
                workspace_id,
                display_name=display_name,
                item_definition=item_definition,
                folder=folder_id,
            )

        else:
            update_dataflow_gen2_definition(
                workspace_id,
                item_id,
                item_definition=item_definition,
            )

    logger.success(
        f'All dataflows_gen2 were deployed to workspace "{workspace}" successfully.'
    )
    return None


def extract_dataflow_gen2_variables(path: str) -> List[Dict[str, Any]]:
    """
    Extract variables from a Dataflow Gen2 mashup.pq file, identifying each destination separately.

    Args:
        path (str): Path to the Dataflow gen2

    Returns:
        (List[Dict[str, Any]]): List of dictionaries containing the extracted variables for each destination
    """
    path = Path(path) / 'mashup.pq'

    with open(path, 'r', encoding='utf-8') as f:
        content = f.read()

    variables = []

    # Pattern to find DataDestination sections
    # Look for shared QueryName_DataDestination = let
    destination_pattern = (
        r'shared\s+(\w+)_DataDestination\s*=\s*let(.*?)in\s*\w+;'
    )
    destination_matches = re.findall(destination_pattern, content, re.DOTALL)

    for query_name, destination_content in destination_matches:
        param_dict = {
            'destination_name': f'{query_name}_DataDestination',
            'query_name': query_name,
        }

        # Extract workspaceId
        workspace_pattern = r'workspaceId\s*=\s*"([a-f0-9-]+)"'
        workspace_match = re.search(workspace_pattern, destination_content)
        if workspace_match:
            param_dict['workspaceId'] = workspace_match.group(1)

        # Extract lakehouseId
        lakehouse_pattern = r'lakehouseId\s*=\s*"([a-f0-9-]+)"'
        lakehouse_match = re.search(lakehouse_pattern, destination_content)
        if lakehouse_match:
            param_dict['lakehouseId'] = lakehouse_match.group(1)
            param_dict['destination_type'] = 'Lakehouse'

        # Extract warehouseId
        warehouse_pattern = r'warehouseId\s*=\s*"([a-f0-9-]+)"'
        warehouse_match = re.search(warehouse_pattern, destination_content)
        if warehouse_match:
            param_dict['warehouseId'] = warehouse_match.group(1)
            param_dict['destination_type'] = 'Warehouse'

        # Extract semanticModelId (if present)
        semantic_pattern = r'semanticModelId\s*=\s*"([a-f0-9-]+)"'
        semantic_match = re.search(semantic_pattern, destination_content)
        if semantic_match:
            param_dict['semanticModelId'] = semantic_match.group(1)
            param_dict['destination_type'] = 'SemanticModel'

        # Only add if we found at least one ID parameter
        if any(key.endswith('Id') for key in param_dict.keys()):
            variables.append(param_dict)

    return variables


def replace_dataflow_gen2_variables_with_placeholders(
    path: str, variables: List[Dict[str, Any]]
) -> None:
    """
    Replace variables with placeholders in a Dataflow Gen2 mashup.pq file.
    Each destination gets unique placeholders based on its query name.

    Args:
        path (str): Path to the Dataflow gen2
        variables (list): List of variable dictionaries to replace

    Returns:
        str: Modified content with placeholders
    """
    dataflow_name = path.split('/')[-1].split('.Dataflow')[0]

    path = Path(path) / 'mashup.pq'

    with open(path, 'r', encoding='utf-8') as f:
        content = f.read()
    # Replace each destination's variables with unique placeholders
    for var_dict in variables:
        query_name = var_dict['query_name']

        # Replace workspaceId
        if 'workspaceId' in var_dict:
            workspace_id = var_dict['workspaceId']
            placeholder = f'#{{{dataflow_name}_{query_name}_workspaceId}}#'
            content = content.replace(
                f'workspaceId = "{workspace_id}"',
                f'workspaceId = "{placeholder}"',
            )

        # Replace lakehouseId
        if 'lakehouseId' in var_dict:
            lakehouse_id = var_dict['lakehouseId']
            placeholder = f'#{{{dataflow_name}_{query_name}_lakehouseId}}#'
            content = content.replace(
                f'lakehouseId = "{lakehouse_id}"',
                f'lakehouseId = "{placeholder}"',
            )

        # Replace warehouseId
        if 'warehouseId' in var_dict:
            warehouse_id = var_dict['warehouseId']
            placeholder = f'#{{{dataflow_name}_{query_name}_warehouseId}}#'
            content = content.replace(
                f'warehouseId = "{warehouse_id}"',
                f'warehouseId = "{placeholder}"',
            )

        # Replace semanticModelId
        if 'semanticModelId' in var_dict:
            semantic_id = var_dict['semanticModelId']
            placeholder = f'#{{{dataflow_name}_{query_name}_semanticModelId}}#'
            content = content.replace(
                f'semanticModelId = "{semantic_id}"',
                f'semanticModelId = "{placeholder}"',
            )

    # Save the modified content back to the file
    with open(path, 'w', encoding='utf-8') as file:
        file.write(content)


def replace_dataflow_gen2_placeholders_with_parameters(
    path: str, variables: List[Dict[str, Any]]
) -> None:
    """
    Replace placeholders with actual parameters in a Dataflow Gen2 mashup.pq file.

    Args:
        path (str): Path to the dataflow gen2
        variables (list): List of variable dictionaries with actual values

    Returns:
        str: Modified content with actual parameter values
    """
    dataflow_name = path.split('/')[-1].split('.Dataflow')[0]

    path = Path(path) / 'mashup.pq'

    with open(path, 'r', encoding='utf-8') as f:
        content = f.read()

    # Replace placeholders with actual values for each destination
    for var_dict in variables:
        query_name = var_dict['query_name']

        # Replace workspaceId placeholder
        if 'workspaceId' in var_dict:
            workspace_id = var_dict['workspaceId']
            placeholder = f'#{{{dataflow_name}_{query_name}_workspaceId}}#'
            content = content.replace(
                f'workspaceId = "{placeholder}"',
                f'workspaceId = "{workspace_id}"',
            )

        # Replace lakehouseId placeholder
        if 'lakehouseId' in var_dict:
            lakehouse_id = var_dict['lakehouseId']
            placeholder = f'#{{{dataflow_name}_{query_name}_lakehouseId}}#'
            content = content.replace(
                f'lakehouseId = "{placeholder}"',
                f'lakehouseId = "{lakehouse_id}"',
            )

        # Replace warehouseId placeholder
        if 'warehouseId' in var_dict:
            warehouse_id = var_dict['warehouseId']
            placeholder = f'#{{{dataflow_name}_{query_name}_warehouseId}}#'
            content = content.replace(
                f'warehouseId = "{placeholder}"',
                f'warehouseId = "{warehouse_id}"',
            )

        # Replace semanticModelId placeholder
        if 'semanticModelId' in var_dict:
            semantic_id = var_dict['semanticModelId']
            placeholder = f'#{{{dataflow_name}_{query_name}_semanticModelId}}#'
            content = content.replace(
                f'semanticModelId = "{placeholder}"',
                f'semanticModelId = "{semantic_id}"',
            )

    with open(path, 'w', encoding='utf-8') as f:
        f.write(content)



import json
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import pandas as pd
from pandas import DataFrame

from ..core.workspaces import resolve_workspace
from ..helpers.folders import resolve_folder_from_id_to_path
from ..items.warehouses import get_warehouse, list_warehouses
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.schemas import PLATFORM_SCHEMA, PLATFORM_VERSION

logger = get_logger(__name__)


def _save_warehouse_sqlproj(
    display_name: str,
    path: Union[Path, str],
) -> None:
    """
    Create a dummy warehouse `.sqlproj` file
    """
    WAREHOUSE_SQL_PROJECT = r"""<Project DefaultTargets="Build">
  <Sdk Name="Microsoft.Build.Sql" Version="0.1.19-preview" />
  <PropertyGroup>
    <Name>{warehouse_display_name}</Name>
    <DSP>Microsoft.Data.Tools.Schema.Sql.SqlDwUnifiedDatabaseSchemaProvider</DSP>
    <DefaultCollation>Latin1_General_100_BIN2_UTF8</DefaultCollation>
  </PropertyGroup>
  <Target Name="BeforeBuild">
    <Delete Files="$(BaseIntermediateOutputPath)\project.assets.json" />
  </Target>
</Project>"""

    sql_project = WAREHOUSE_SQL_PROJECT.format(
        warehouse_display_name=display_name
    )

    with open(Path(path) / f'{display_name}.sqlproj', 'w') as f:
        f.write(sql_project)

    logger.success(
        f'{display_name}.sqlproject has been created in {path} successfully.'
    )


def _save_warehouse_defaultsemanticmodel_txt(
    path: Union[Path, str],
) -> None:
    """
    Create a `DefaultSemanticModel.txt` in `Warehouse` path.
    """
    with open(Path(path) / 'DefaultSemanticModel.txt', 'w') as f:
        f.write('Has default semantic model')

    logger.success(
        f'DefaultSemanticModel.txt was created in {path} successfully.'
    )


def _save_warehouse_xmla_json(
    path: Union[Path, str],
) -> None:
    """
    Create a dummy `xmla.json` on `Warehouse` path.
    """
    WAREHOUSE_XMLA_JSON = {
        'name': '{{Dataset_Name}}',
        'compatibilityLevel': 1604,
        'model': {
            'name': '{{Dataset_Name}}',
            'culture': 'en-US',
            'collation': 'Latin1_General_100_BIN2_UTF8',
            'dataAccessOptions': {
                'legacyRedirects': True,
                'returnErrorValuesAsNull': True,
            },
            'defaultPowerBIDataSourceVersion': 'powerBI_V3',
            'sourceQueryCulture': 'en-US',
            'expressions': [
                {
                    'name': 'DatabaseQuery',
                    'kind': 'm',
                    'expression': 'let\n    database = {{TDS_Endpoint}}\nin\n    database\n',
                }
            ],
            'annotations': [
                {'name': '__PBI_TimeIntelligenceEnabled', 'value': '0'},
                {
                    'name': 'SourceLineageTagType',
                    'value': 'DatabaseFullyQualifiedName',
                },
            ],
        },
    }
    with open(Path(path) / 'xmla.json', 'w') as f:
        json.dump(WAREHOUSE_XMLA_JSON, f, indent=2)

    logger.success(f'xmla.json was created in {path} successfully.')


def _generate_warehouse_platform(
    display_name: str,
    description: Optional[str] = '',
) -> Dict[str, Any]:
    """
    Generate the warehouse .platform file

    Args:
        display_name (str): The warehouse display name.
        description (str): The warehouse's description.

    Returns:
        (Dict[str, Any]): The .platform dict.
    """
    return {
        '$schema': PLATFORM_SCHEMA,
        'metadata': {
            'type': 'Warehouse',
            'displayName': display_name,
            'description': description,
        },
        'config': {
            'version': PLATFORM_VERSION,
            'logicalId': '00000000-0000-0000-0000-000000000000',
        },
    }


def _save_warehouse_platform(
    platform: Dict[str, Any],
    path: str,
) -> None:
    """
    Save the warehouses's .platform in path

    Args:
        platform (Dict[str, Any]): The .platform dict.
        path (str): The warehouse directory path to save to.
    """
    with open(Path(path) / '.platform', 'w') as f:
        json.dump(platform, f, indent=2)


def get_warehouse_config(
    workspace: str, warehouse: str
) -> Union[Dict[str, Any], None]:
    """
    Get a specific warehouse config from a workspace.

    Args:
        workspace (str): The name or ID from the workspace.
        warehouse (str): The name or ID from the warehouse.

    Returns:
        (Union[Dict[str, Any], None]): The dict config from the warehouse
    """
    item = warehouse
    item_data = get_warehouse(workspace, item, df=False)

    if item_data is None:
        return None

    else:
        config = {}
        config = config[item_data.get('displayName')] = {}

        config = {
            'id': item_data['id'],
            'description': item_data.get('description', None),
            'folder_id': ''
            if item_data.get('folderId') is None
            or pd.isna(item_data.get('folderId'))
            else item_data['folderId'],
            'connection_string': item_data['properties']['connectionString'],
        }

        return config


def get_all_warehouses_config(workspace: str) -> Union[Dict[str, Any], None]:
    """
    Generate warehouses config from a workspace.

    Args:
        workspace (str): The name or ID from the workspace.

    Returns:
        (Union[Dict[str, Any], None]): The dict config from the warehouses of the workspace
    """
    items = list_valid_warehouses(workspace, df=False)

    if items is None:
        return None

    config = {}

    for item in items:

        item_data = get_warehouse(workspace, item['id'], df=False)

        config[item['displayName']] = {
            'id': item['id'],
            'description': item.get('description', None),
            'folder_id': ''
            if item.get('folderId') is None or pd.isna(item.get('folderId'))
            else item['folderId'],
            'connection_string': item_data['properties']['connectionString'],
        }

    return config


@df
def list_valid_warehouses(
    workspace: str,
    df: Optional[bool] = True,
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Generate a list of valid warehouses from a workspace.

    Args:
        workspace (str): The name or ID from the workspace.

    Returns:
        (Union[Dict[str, Any], None]): The list of valids warehouses of the workspace
    """
    items = list_warehouses(workspace)

    if items is None:
        return None

    return items[
        ~items['displayName'].str.contains('staging', case=False, na=False)
    ].to_dict(orient='records')


def export_warehouse(
    workspace: str,
    warehouse: str,
    path: Union[str, Path],
) -> None:
    """
    Export a warehouse to path

    Args:
        workspace (str): The name or ID of the workspace.
        warehouse (str): The name or ID of the warehouse.
        path (Union[str, Path]): The path to export to.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    item = get_warehouse(workspace_id, warehouse, df=True)
    try:
        folder_path = resolve_folder_from_id_to_path(
            workspace_id, item['folderId']
        )
    except:
        logger.info(f'{item["displayName"]}.Warehouse is not inside a folder.')
        folder_path = None

    if folder_path is None:
        item_path = Path(path) / (item['displayName'] + '.Warehouse')
    else:
        item_path = (
            Path(path) / folder_path / (item['displayName'] + '.Warehouse')
        )
    os.makedirs(item_path, exist_ok=True)

    platform = _generate_warehouse_platform(
        display_name=item['displayName'],
        description=item['description'],
    )

    _save_warehouse_platform(platform, item_path)

    _save_warehouse_defaultsemanticmodel_txt(item_path)

    _save_warehouse_sqlproj(item['displayName'], item_path)

    _save_warehouse_xmla_json(item_path)

    logger.success(f'All warehouses exported to {path} successfully.')
    return None


def export_all_warehouses(workspace: str, path: Union[str, Path]) -> None:
    """
    Exports all warehouses from the workspace to path.

    Args:
        workspace (str): The ID or name of the workspace.
        path (Union[str, Path]): The path to export to.
    """
    workspace_id = resolve_workspace(workspace)
    if workspace_id is None:
        return None

    items = list_valid_warehouses(workspace_id, df=False)
    if items is None:
        return None

    for item in items:
        try:
            folder_path = resolve_folder_from_id_to_path(
                workspace_id, item['folderId']
            )
        except:
            logger.info(
                f'{item["displayName"]}.Warehouse is not inside a folder.'
            )
            folder_path = None

        if folder_path is None:
            item_path = Path(path) / (item['displayName'] + '.Warehouse')
        else:
            item_path = (
                Path(path) / folder_path / (item['displayName'] + '.Warehouse')
            )
        os.makedirs(item_path, exist_ok=True)

        platform = _generate_warehouse_platform(
            display_name=item['displayName'],
            description=item['description'],
        )

        _save_warehouse_platform(platform, item_path)

        _save_warehouse_defaultsemanticmodel_txt(item_path)

        _save_warehouse_sqlproj(item['displayName'], item_path)

        _save_warehouse_xmla_json(item_path)

    logger.success(f'All warehouses exported to {path} successfully.')
    return None



from typing import Any, Dict, List, Optional, Union

from pandas import DataFrame

from ..api.api import api_request
from ..core.folders import resolve_folder
from ..core.workspaces import resolve_workspace
from ..utils.decorators import df
from ..utils.logging import get_logger
from ..utils.utils import is_valid_uuid

logger = get_logger(__name__)


@df
def list_data_pipelines(
    workspace: str, *, df: Optional[bool] = True
) -> Union[DataFrame, List[Dict[str, Any]], None]:
    """
    Lists all data_pipelines in the specified workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        list | pandas.DataFrame | None: A list of data_pipelines if successful, otherwise None.

    Examples:
        ```python
        list_data_pipelines('MyProjectWorkspace')
        list_data_pipelines('123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    return api_request(
        '/workspaces/' + resolve_workspace(workspace) + '/dataPipelines',
        support_pagination=True,
    )


def get_data_pipeline_id(
    workspace: str, data_pipeline_name: str
) -> Union[str, None]:
    """
    Retrieves the ID of a data pipeline by its name.

    Args:
        data_pipeline_name (str): The name of the data pipeline.

    Returns:
        str | None: The ID of the data pipeline if found, otherwise None.
    """
    data_pipelines = list_data_pipelines(
        workspace=resolve_workspace(workspace),
        df=False,
    )
    for _data_pipeline in data_pipelines:
        if _data_pipeline['displayName'] == data_pipeline_name:
            return _data_pipeline['id']
    logger.warning(
        f"DataPipeline '{data_pipeline_name}' not found in workspace '{workspace}'."
    )
    return None


def resolve_data_pipeline(
    workspace: str, data_pipeline: str
) -> Union[str, None]:
    """
    Resolves a data pipeline name to its ID.

    Args:
        workspace (str): The name or ID of the workspace.
        data_pipeline (str): The name or ID of the data pipeline.

    Returns:
        str | None: The ID of the data pipeline if found, otherwise None.
    """
    if is_valid_uuid(data_pipeline):
        return data_pipeline
    else:
        return get_data_pipeline_id(workspace, data_pipeline)


@df
def get_data_pipeline(
    workspace: str, data_pipeline: str
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Retrieves the details of a data pipeline by its ID.

    Args:
        workspace (str): The name or ID of the workspace.
        data_pipeline (str): The name or ID of the data pipeline.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.


    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The data pipeline details if found, otherwise None.

    Examples:
        ```python
        get_data_pipeline('123e4567-e89b-12d3-a456-426614174000', 'SalesDataPipeline')
        get_data_pipeline('my-workspace', 'SalesDataPipeline')
        ```
    """
    workspace_id = resolve_workspace(workspace)
    return api_request(
        '/workspaces/'
        + workspace_id
        + '/dataPipelines/'
        + resolve_data_pipeline(workspace_id, data_pipeline),
    )


@df
def update_data_pipeline(
    workspace: str,
    data_pipeline: str,
    *,
    display_name: Optional[str] = None,
    description: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates the properties of the specified data pipeline.

    Args:
        workspace (str): The workspace name or ID.
        data_pipeline (str): The name or ID of the data_pipeline to update.
        display_name (str, optional): The new display name for the data_pipeline.
        description (str, optional): The new description for the data_pipeline.
        df (bool, optional): Keyword-only. If True, returns a DataFrame with flattened keys. Defaults to False.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated data pipeline details if successful, otherwise None.

    Examples:
        ```python
        update_data_pipeline('MyProjectWorkspace', 'SalesDataModel', display_name='UpdatedSalesDataModel')
        update_data_pipeline('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000', description='Updated description')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    data_pipeline_id = resolve_data_pipeline(workspace_id, data_pipeline)

    payload = {}

    if display_name:
        payload['displayName'] = display_name

    if description:
        payload['description'] = description

    return api_request(
        '/workspaces/' + workspace_id + '/dataPipelines/' + data_pipeline_id,
        method='patch',
        payload=payload,
    )


def delete_data_pipeline(workspace: str, data_pipeline: str) -> None:
    """
    Delete a data_pipeline from the specified workspace.

    Args:
        workspace (str): The name or ID of the workspace to delete.
        data_pipeline (str): The name or ID of the data_pipeline to delete.

    Returns:
        None: If the data_pipeline is successfully deleted.

    Raises:
        ResourceNotFoundError: If the specified workspace is not found.

    Examples:
        ```python
        delete_data_pipeline('123e4567-e89b-12d3-a456-426614174000', 'Salesdata_pipeline')
        delete_data_pipeline('MyProject', 'Salesdata_pipeline')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    data_pipeline_id = resolve_data_pipeline(workspace_id, data_pipeline)

    return api_request(
        '/workspaces/' + workspace_id + '/dataPipelines/' + data_pipeline_id,
        method='delete',
    )


def get_data_pipeline_definition(workspace: str, data_pipeline: str) -> dict:
    """
    Retrieves the definition of a data_pipeline by its name or ID from the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        data_pipeline (str): The name or ID of the data_pipeline.

    Returns:
        (dict): The data_pipeline definition if found, otherwise None.

    Examples:
        ```python
        get_data_pipeline_definition('MyProjectWorkspace', 'Salesdata_pipeline')
        get_data_pipeline_definition('MyProjectWorkspace', '123e4567-e89b-12d3-a456-426614174000')
        ```
    """
    # Resolving IDs
    workspace_id = resolve_workspace(workspace)

    data_pipeline_id = resolve_data_pipeline(workspace_id, data_pipeline)

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/dataPipelines/'
        + data_pipeline_id
        + '/getDefinition',
        method='post',
        support_lro=True,
    )


@df
def update_data_pipeline_definition(
    workspace: str, data_pipeline: str, item_definition: Dict[str, Any]
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Updates the definition of an existing data_pipeline in the specified workspace.
    If the data_pipeline does not exist, it returns None.

    Args:
        workspace (str): The workspace name or ID.
        data_pipeline (str): The name or ID of the data_pipeline to update.
        item_definition (Dict[str, Any]): The item_definition of the data_pipeline.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The updated data_pipeline details if successful, otherwise None.

    Examples:
        ```python
        update_data_pipeline_definition(
            workspace='MyProjectWorkspace',
            data_pipeline='SalesDataPipeline',
            item_definition={...} # The definition of the data_pipeline
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    data_pipeline_id = resolve_data_pipeline(workspace_id, data_pipeline)

    params = {'updateMetadata': True}
    payload = {'definition': item_definition}

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/dataPipelines/'
        + data_pipeline_id,
        method='post',
        payload=payload,
        params=params,
    )


@df
def create_data_pipeline(
    workspace: str,
    display_name: str,
    item_definition: Dict[str, Any],
    *,
    description: Optional[str] = None,
    folder: Optional[str] = None,
    df: Optional[bool] = True,
) -> Union[DataFrame, Dict[str, Any], None]:
    """
    Creates a new data_pipeline in the specified workspace.

    Args:
        workspace (str): The workspace name or ID.
        display_name (str): The display name of the data_pipeline.
        description (str, optional): A description for the data_pipeline.
        folder (str, optional): The folder to create the data_pipeline in.
        item_definition (Dict[str, Any]): The definition of the data_pipeline.
        df (Optional[bool]): If True or not provided, returns a DataFrame with flattened keys.
            If False, returns a list of dictionaries.

    Returns:
        (Union[DataFrame, Dict[str, Any], None]): The created data_pipeline details.

    Examples:
        ```python
        create_data_pipeline(
            workspace='MyProjectWorkspace',
            display_name='SalesDataPipeline',
            item_definition={...},
            description='This is a sales data pipeline',
            folder='SalesDataPipelinesFolder'
        )
        ```
    """
    workspace_id = resolve_workspace(workspace)

    payload = {
        'displayName': display_name,
        'definition': item_definition,
    }

    if folder:
        folder_id = resolve_folder(workspace_id, folder)
        if folder_id:
            payload['folderId'] = folder_id

    if description:
        payload['description'] = description

    return api_request(
        endpoint='/workspaces/' + workspace_id + '/dataPipelines',
        method='post',
        payload=payload,
    )


def delete_data_pipeline(workspace: str, data_pipeline: str) -> None:
    """
    Deletes a data_pipeline from the specified workspace.

    Args:
        workspace (str): The name or ID of the workspace.
        data_pipeline (str): The name or ID of the data_pipeline to delete.

    Returns:
        None: If the data_pipeline is successfully deleted.

    Examples:
        ```python
        delete_data_pipeline('MyProjectWorkspace', 'SalesDataPipeline')
        delete_data_pipeline('123e4567-e89b-12d3-a456-426614174000', 'SalesDataPipeline')
        ```
    """
    workspace_id = resolve_workspace(workspace)

    data_pipeline_id = resolve_data_pipeline(workspace_id, data_pipeline)

    return api_request(
        endpoint='/workspaces/'
        + workspace_id
        + '/dataPipelines/'
        + data_pipeline_id,
        method='delete',
    )
